import re
from urllib.parse import urljoin

from ready import thttp
from ready.checks.csp import extract_csp
from ready.result import result

USE_BS4 = True

try:
    from bs4 import BeautifulSoup
except ImportError:
    USE_BS4 = False


# Check: Permissions-Policy should exist if the response is HTML
def check_permissions_policy_should_exist(responses, **kwargs):
    return result(
        responses["response"].headers.get("permissions-policy") != None,
        f"Permissions-Policy should exist if the response is HTML ({responses['response'].headers.get('permissions-policy')})",
        "html_permissions_policy",
        **kwargs,
    )


# Check: frame-ancestors should be in CSP or X-Frame-Options should exist if the response is HTML
def check_frame_ancestors_should_exist(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        responses["response"].headers.get("x-frame-options") != None or (csp != None and "frame-ancestors" in csp),
        f"frame-ancestors should be in CSP or X-Frame-Options should exist if the response is HTML (X-Frame-Options: {responses['response'].headers.get('x-frame-options')}, CSP: {csp})",
        "html_frame_ancestors",
        **kwargs,
    )


# Check: X-Content-Type-options should be "nosniff"
def check_x_content_type_options_should_be_nosniff(responses, **kwargs):
    return result(
        responses["response"].headers.get("x-content-type-options", "") == "nosniff",
        f'X-Content-Type-options should be "nosniff" ({responses["response"].headers.get("x-content-type-options", "")})',
        "html_x_content_type_options_nosniff",
        **kwargs,
    )


# Check: Referrer-Policy should be set
def check_referrer_policy_should_be_set(responses, **kwargs):
    return result(
        responses["response"].headers.get("referrer-policy") != None,
        f'Referrer-Policy should be set ({responses["response"].headers.get("referrer-policy", "")})',
        "html_referrer_policy",
        **kwargs,
    )


# Check: X-XSS-Protection header should not exist
def check_x_xss_protection_should_not_exist(responses, **kwargs):
    return result(
        "x-xss-protection" not in responses["response"].headers,
        f'X-XSS-Protection header should not exist ({responses["response"].headers.get("x-xss-protection")})',
        "html_x_xss_protection_not_set",
        warn_on_fail=True,
        **kwargs,
    )


# Check: HTML should start with "<!doctype html>"
def check_html_starts_with_doctype(responses, **kwargs):
    return result(
        responses["response"].content.lower().strip().startswith(b"<!doctype html>"),
        f'HTML should start with "<!doctype html>"',
        "html_doctype",
        **kwargs,
    )


# Check: `<html>` tag should include lang
def check_html_tag_includes_lang(responses, **kwargs):
    if "<html" in responses["response"].content.decode(errors="ignore"):
        html_tag = responses["response"].content.decode(errors="ignore").split("<html")[1].split(">")[0].replace("'", '"')
        html_tag = "<html" + html_tag + ">"
    else:
        html_tag = "no tag"
    return result(
        "lang=" in html_tag,
        f"<html> tag should include lang ({html_tag})",
        "html_tag_includes_lang",
        **kwargs,
    )


# Check: HTML should include meta charset tag
def check_html_meta_charset(responses, **kwargs):
    return result(
        b"<meta charset=" in responses["response"].content.lower(),
        f"HTML should include meta charset tag",
        "html_meta_charset",
        **kwargs,
    )


# Check: HTML should include `<title>`
def check_html_includes_title(responses, **kwargs):
    return result(
        b"<title>" in responses["response"].content.lower(),
        f"HTML should include title",
        "html_includes_title",
        **kwargs,
    )


# Check: HTML should include link with rel="icon"
def check_html_includes_rel_icon(responses, **kwargs):
    link_re = re.compile(b"<link (.+)>")
    links = [l.replace(b"'", b'"') for l in link_re.findall(responses["response"].content)]

    return result(
        any([b'rel="icon"' in link for link in links]) or any([b'rel="shortcut icon"' in link for link in links]),
        'HTML should include link with rel="icon"',
        "html_rel_icon",
        **kwargs,
    )


# Check: HTML should not use schemeless urls for links or hrefs
def check_html_should_not_use_schemeless_urls(responses, **kwargs):
    return result(
        b'="//' not in responses["response"].content and b"='//" not in responses["response"].content,
        "HTML should not use schemeless urls for links or hrefs",
        "html_schemeless",
        **kwargs,
    )


# Check: HTML should not use unnecessary HTML entities
def check_html_should_not_use_unnecessary_entities(responses, **kwargs):
    allow_list = [b"nbsp", b"amp", b"quot", b"lt", b"gt"]

    # The longest entity on the registered entity list is "CounterClockwiseContourIntegral"
    # https://html.spec.whatwg.org/entities.json
    entities = re.findall(b"&([\w#]{1,32});", responses["response"].content)
    entities = [e for e in entities if e not in allow_list]

    return result(
        len(entities) == 0,
        f"HTML should not use unnecessary HTML entities ({[e.decode() for e in set(entities)]})",
        "html_unnecessary_entities",
        warn_on_fail=True,
        **kwargs,
    )
    print(entities)


# Check: All script tags should use subresource integrity
def check_html_script_tags_use_sri(responses, **kwargs):
    script_tags = re.findall(b"<script ([^\>]+)", responses["response"].content)

    return result(
        all([b"integrity" in tag for tag in script_tags]),
        f"All script tags should use subresource integrity",
        "html_sri_js",
        **kwargs,
    )


# Check: X-DNS-Prefetch-Control should be set to off
def check_x_dns_prefetch_control_is_off(responses, **kwargs):
    return result(
        responses["response"].headers.get("x-dns-prefetch-control", "") == "off",
        f"X-DNS-Prefetch-Control should be set to off ({responses['response'].headers.get('x-dns-prefetch-control', '')})",
        "html_x_dns_prefetch",
        **kwargs,
    )


# Check: CDNs should not be used for Javascript or CSS assets
def check_cdns_should_not_be_used(responses, **kwargs):
    # XXX: This list was compiled by myself from a number of random web sources, if a better maintained list
    # exists then I would love to replace this
    cdn_domains = [
        "cdn.jsdelivr.net",
        "cdn.statically.io",
        "bootstrapcdn.com",
        "cdnjs.cloudflare.com",
        "sentry-cdn.com",
        "ajax.googleapis.com",
        "code.jquery.com",
        "ajax.aspnetcdn.com",
    ]

    script_tags = re.findall(b"<script ([^\>]+)", responses["response"].content)
    link_tags = re.findall(b"<link (.+)>", responses["response"].content)

    for tag in script_tags + link_tags:
        if any([x in tag.decode() for x in cdn_domains]):
            return result(False, "CDNs should not be used for Javascript or CSS assets", "html_cdn_usage", **kwargs)

    return result(True, "CDNs should not be used for Javascript or CSS assets", "html_cdn_usage", **kwargs)


# Check: RSS and JSON feeds should return Access-Control-Allow-Origin header
def check_rss_should_return_cors_header(responses, **kwargs):
    if USE_BS4:
        feed_urls = []
        feed_types = [
            "application/rss+xml",
            "application/feed+json",
        ]

        soup = BeautifulSoup(responses["response"].content, features="html.parser")

        links = soup.find_all("link")

        for link in links:
            if "alternate" in link.attrs.get("rel", ""):
                if link.attrs.get("type", "") in feed_types:
                    feed_urls.append(urljoin(responses["response"].url, link.attrs.get("href")))

        cors_values = []
        for url in feed_urls:
            if url.startswith("//"):
                url = "https:" + url
            elif url.startswith("/"):
                url = responses["response"].url.rstrip("/") + url

            if url.startswith("http"):
                response = thttp.request(url)
                cors_values.append(response.headers.get("access-control-allow-origin"))

        return result(
            all([x is not None for x in cors_values]),
            f"RSS and JSON feeds should return Access-Control-Allow-Origin header ({', '.join(feed_urls) if feed_urls else 'no feeds'})",
            "feeds_cors_enabled",
            **kwargs,
        )
    else:
        return result(
            False,
            f"RSS and JSON feeds should return Access-Control-Allow-Origin header (skipped because beautifulsoup is missing)",
            "feeds_cors_enabled",
            warn_on_fail=True,
            **kwargs,
        )


# Check: Cache-Control max-age should be <= 86400 for HTML documents
def check_html_should_not_be_cached_for_more_than_24_hours(responses, **kwargs):
    cc_header = responses["response"].headers.get("cache-control", "")
    error = "no Cache-Control header"

    if "max-age=" in cc_header:
        max_age = re.search("max-age=(?P<age>\d+)", cc_header)

        if max_age:
            try:
                age = int(max_age.group("age"))
                return result(
                    age <= 86400,
                    f"Cache-Control max-age should be <= 86400 for HTML documents ({cc_header})",
                    "html_cache_duration",
                    **kwargs,
                )
            except ValueError:
                error = f"parse error: {cc_header}"
        else:
            error = f"match error: {cc_header}"

    return result(
        False,
        f"Cache-Control max-age should be <= 86400 for HTML documents ({error})",
        "html_cache_duration",
        warn_on_fail=True,
        **kwargs,
    )



from unittest import TestCase

from ready.checks.csp import extract_csp
from ready.thttp import Response


class ReadyTestCase(TestCase):
    def test_extract_csp_meta_tag(self):
        content = """<!doctype html>
        <html lang="en">
        <head>
            <meta charset="utf-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <meta name="description" content="">
            <meta
              http-equiv="Content-Security-Policy"
              content="default-src 'self'; img-src https://*; child-src 'none';" />

            <title>Minimal base.html</title>
        </head>
        <body>

            <!-- Delete this part -->
            <h1>base.html</h1>
            <p>The absolute minimum <code>base.html</code> to get your project started.</p>
            <h3>Usage:</h3>
            <pre>
              curl https://basehtml.xyz &gt; base.html
            </pre>
            <a href="https://github.com/sesh/base.html">more info</a>
        </body>
        </html>"""

        # request content json status url headers cookiejar
        response = Response(None, content, None, None, None, {}, None)

        try:
            from bs4 import BeautifulSoup

            csp = extract_csp(response)
            self.assertEqual(csp, "default-src 'self'; img-src https://*; child-src 'none';")
        except ImportError:
            print("Not executing CSP test because beautifulsoup is not available.")



class ReadyResult:
    def __init__(self, passed, message, check, warn_on_fail, *, domain=None):
        self.passed = passed
        self.message = message
        self.check = check
        self.warn_on_fail = warn_on_fail
        self.domain = domain


def result(
    passed,
    message,
    check,
    *,
    warn_on_fail=False,
    colour=True,
    print_output=True,
    **kwargs,
):
    RED = "\033[0;31m" if colour else ""
    GREEN = "\033[0;32m" if colour else ""
    YELLOW = "\033[0;33m" if colour else ""
    NC = "\033[0m" if colour else ""  # No Color

    pre = f"[ {GREEN}OK{NC} ]"
    if not passed and not warn_on_fail:
        pre = f"[{RED}FAIL{NC}]"
    elif not passed:
        pre = f"[{YELLOW}WARN{NC}]"

    if print_output:
        print(f"{pre} {message}")

    return ReadyResult(passed, message, check, warn_on_fail, domain=kwargs.get("domain"))



import re
from urllib.parse import urljoin

from ready.result import result
from ready.thttp import request

SWAGGER_PATHS = [
    "core/latest/swagger-ui/index.html",
    "csp/gateway/slc/api/swagger-ui.html",
    "swagger",
    "swagger-resources",
    "swagger-ui",
    "swagger-ui.html",
    "swagger.json",
    "swagger.yaml",
    "swagger/index.html",
    "swagger/swagger-ui.htm",
    "swagger/swagger-ui.html",
    "swagger/ui",
    "swagger/v1/swagger.json",
    "swaggerui",
]


# Check: Swagger URLs should not return 200 (requires --fuzz)
def check_swagger_should_not_return_200(responses, **kwargs):
    url = responses["response"].url

    swagger_responses = []
    for path in SWAGGER_PATHS:
        response = request(urljoin(url, path))
        if response.status < 299:
            swagger_responses.append(response)

    return result(
        len(swagger_responses) == 0,
        f"Swagger URLs should not return 200 ({[(r.url, r.status) for r in swagger_responses]})",
        "cors_header_exists",
        **kwargs,
    )



import json

from ready.result import result


# Check: An AAAA DNS record exists (IPv6 Support)
def check_aaaa_record_exists(responses, **kwargs):
    records = [x["data"] for x in responses["dns_aaaa_response"].json.get("Answer", [])]

    return result(
        len(records) >= 1,
        f"An AAAA DNS record exists ({records})",
        "dns_aaaa",
        **kwargs,
    )



import re

from ready.result import result

LEAKY_HEADERS = [
    "apigw-requestid",
    "cdn-cache",
    "cf-edge-cache",
    "fastly-debug-states",
    "fly-request-id",
    "ghost-fastly",
    "served-by",
    "server",
    "x-appversion",
    "x-aspnet-version",
    "x-aspnetmvc-version",
    "x-backend-name",
    "x-backend-server",
    "x-backend",
    "x-build-id",
    "x-build",
    "x-cache-info",
    "x-cache-key",
    "x-cache-rule",
    "x-cached-by",
    "x-cdn-rule",
    "x-cdn",
    "x-cf-worker",
    "x-client-ip",
    "x-diaspora-version",
    "x-drupal-theme",
    "x-fastly-request-id",
    "x-fw-version",
    "x-generator",
    "x-github-backend",
    "x-hosted-by",
    "x-httpd",
    "x-kinja-revision",
    "x-lambda-id",
    "x-last-commmit-hash",
    "x-litespeed-cache",
    "x-nextjs-page",
    "x-nodejs",
    "x-origin-server",
    "x-powered-by-plesk",
    "x-powered-by",
    "x-powered",
    "x-protected-by",
    "x-provided-by",
    "x-section",
    "x-server-powered-by",
    "x-server",
    "x-tumblr-user",
    "x-varnish",
    "x-vercel-id",
    "x-version",
    "via",
]


# Check: Headers that leak information should not be in the response
def check_should_not_include_leaky_headers(responses, **kwargs):
    leaky = [x for x in LEAKY_HEADERS if re.search(r"\d\.\d", responses["response"].headers.get(x, "")) != None]
    return result(
        len(leaky) == 0,
        f"Headers that leak information should not be in the response ({leaky})",
        "leaky_headers",
        **kwargs,
    )



"""
UNLICENSED
This is free and unencumbered software released into the public domain.

https://github.com/sesh/thttp
"""

import gzip
import json as json_lib
import ssl
from base64 import b64encode
from collections import namedtuple
from http.cookiejar import CookieJar
from urllib.error import HTTPError, URLError
from urllib.parse import urlencode
from urllib.request import (
    HTTPCookieProcessor,
    HTTPRedirectHandler,
    HTTPSHandler,
    Request,
    build_opener,
)

Response = namedtuple("Response", "request content json status url headers cookiejar")


JSON_HEADERS = ["application/x-javascript", "application/json"]


class NoRedirect(HTTPRedirectHandler):
    def redirect_request(self, req, fp, code, msg, headers, newurl):
        return None


def request(
    url,
    params={},
    json=None,
    data=None,
    headers={},
    method="GET",
    verify=True,
    redirect=True,
    cookiejar=None,
    basic_auth=None,
    timeout=None,
):
    """
    Returns a (named)tuple with the following properties:
        - request
        - content
        - json (dict; or None)
        - headers (dict; all lowercase keys)
            - https://stackoverflow.com/questions/5258977/are-http-headers-case-sensitive
        - status
        - url (final url, after any redirects)
        - cookiejar
    """
    method = method.upper()
    headers = {k.lower(): v for k, v in headers.items()}  # lowecase headers

    if params:
        url += "?" + urlencode(params)  # build URL from params
    if json and data:
        raise Exception("Cannot provide both json and data parameters")
    if method not in ["POST", "PATCH", "PUT"] and (json or data):
        raise Exception("Request method must POST, PATCH or PUT if json or data is provided")
    if not timeout:
        timeout = 60

    if json:  # if we have json, stringify and put it in our data variable
        headers["content-type"] = "application/json"
        data = json_lib.dumps(json).encode("utf-8")
    elif data:
        data = urlencode(data).encode()

    if basic_auth and len(basic_auth) == 2 and "authorization" not in headers:
        username, password = basic_auth
        headers["authorization"] = f'Basic {b64encode(f"{username}:{password}".encode()).decode("ascii")}'

    if not cookiejar:
        cookiejar = CookieJar()

    ctx = ssl.create_default_context()
    if not verify:  # ignore ssl errors
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

    handlers = []
    handlers.append(HTTPSHandler(context=ctx))
    handlers.append(HTTPCookieProcessor(cookiejar=cookiejar))

    if not redirect:
        no_redirect = NoRedirect()
        handlers.append(no_redirect)

    opener = build_opener(*handlers)
    req = Request(url, data=data, headers=headers, method=method)

    try:
        with opener.open(req, timeout=timeout) as resp:
            status, content, resp_url = (resp.getcode(), resp.read(), resp.geturl())
            headers = {k.lower(): v for k, v in list(resp.info().items())}

            if "gzip" in headers.get("content-encoding", ""):
                content = gzip.decompress(content)

            json = json_lib.loads(content) if any([x in headers.get("content-type", "").lower() for x in JSON_HEADERS]) else None
    except HTTPError as e:
        status, content, resp_url = (e.code, e.read(), e.geturl())
        headers = {k.lower(): v for k, v in list(e.headers.items())}

        if "gzip" in headers.get("content-encoding", ""):
            content = gzip.decompress(content)

        json = json_lib.loads(content) if any([x in headers.get("content-type", "").lower() for x in JSON_HEADERS]) else None

    return Response(req, content, json, status, resp_url, headers, cookiejar)


import unittest


class RequestTestCase(unittest.TestCase):
    def test_cannot_provide_json_and_data(self):
        with self.assertRaises(Exception):
            request(
                "https://httpbingo.org/post",
                json={"name": "Brenton"},
                data="This is some form data",
            )

    def test_should_fail_if_json_or_data_and_not_p_method(self):
        with self.assertRaises(Exception):
            request("https://httpbingo.org/post", json={"name": "Brenton"})

        with self.assertRaises(Exception):
            request("https://httpbingo.org/post", json={"name": "Brenton"}, method="HEAD")

    def test_should_set_content_type_for_json_request(self):
        response = request("https://httpbingo.org/post", json={"name": "Brenton"}, method="POST")
        self.assertEqual(response.request.headers["Content-type"], "application/json")

    def test_should_work(self):
        response = request("https://httpbingo.org/get")
        self.assertEqual(response.status, 200)

    def test_should_create_url_from_params(self):
        response = request(
            "https://httpbingo.org/get",
            params={"name": "brenton", "library": "tiny-request"},
        )
        self.assertEqual(response.url, "https://httpbingo.org/get?name=brenton&library=tiny-request")

    def test_should_return_headers(self):
        response = request("https://httpbingo.org/response-headers", params={"Test-Header": "value"})
        self.assertEqual(response.headers["test-header"], "value")

    def test_should_populate_json(self):
        response = request("https://httpbingo.org/json")
        self.assertTrue("slideshow" in response.json)

    def test_should_return_response_for_404(self):
        response = request("https://httpbingo.org/404")
        self.assertEqual(response.status, 404)
        self.assertTrue("text/plain" in response.headers["content-type"])

    def test_should_fail_with_bad_ssl(self):
        with self.assertRaises(URLError):
            response = request("https://expired.badssl.com/")

    def test_should_load_bad_ssl_with_verify_false(self):
        response = request("https://expired.badssl.com/", verify=False)
        self.assertEqual(response.status, 200)

    def test_should_form_encode_non_json_post_requests(self):
        response = request("https://httpbingo.org/post", data={"name": "test-user"}, method="POST")
        self.assertEqual(response.json["form"]["name"], ["test-user"])

    def test_should_follow_redirect(self):
        response = request(
            "https://httpbingo.org/redirect-to",
            params={"url": "https://duckduckgo.com/"},
        )
        self.assertEqual(response.url, "https://duckduckgo.com/")
        self.assertEqual(response.status, 200)

    def test_should_not_follow_redirect_if_redirect_false(self):
        response = request(
            "https://httpbingo.org/redirect-to",
            params={"url": "https://duckduckgo.com/"},
            redirect=False,
        )
        self.assertEqual(response.status, 302)

    def test_cookies(self):
        response = request(
            "https://httpbingo.org/cookies/set",
            params={"cookie": "test"},
            redirect=False,
        )
        response = request("https://httpbingo.org/cookies", cookiejar=response.cookiejar)
        self.assertEqual(response.json["cookie"], "test")

    def test_basic_auth(self):
        response = request("http://httpbingo.org/basic-auth/user/passwd", basic_auth=("user", "passwd"))
        self.assertEqual(response.json["authorized"], True)

    def test_should_handle_gzip(self):
        response = request("http://httpbingo.org/gzip", headers={"Accept-Encoding": "gzip"})
        self.assertEqual(response.json["gzipped"], True)

    def test_should_timeout(self):
        with self.assertRaises(TimeoutError):
            response = request("http://httpbingo.org/delay/3", timeout=1)


STATUS_CODES = {
    "100": "Continue",
    "101": "Switching Protocols",
    "102": "Processing",
    "103": "Early Hints",
    "200": "OK",
    "201": "Created",
    "202": "Accepted",
    "203": "Non-Authoritative Information",
    "204": "No Content",
    "205": "Reset Content",
    "206": "Partial Content",
    "207": "Multi-Status",
    "208": "Already Reported",
    "226": "IM Used",
    "300": "Multiple Choices",
    "301": "Moved Permanently",
    "302": "Found",
    "303": "See Other",
    "304": "Not Modified",
    "305": "Use Proxy",
    "307": "Temporary Redirect",
    "308": "Permanent Redirect",
    "400": "Bad Request",
    "401": "Unauthorized",
    "402": "Payment Required",
    "403": "Forbidden",
    "404": "Not Found",
    "405": "Method Not Allowed",
    "406": "Not Acceptable",
    "407": "Proxy Authentication Required",
    "408": "Request Timeout",
    "409": "Conflict",
    "410": "Gone",
    "411": "Length Required",
    "412": "Precondition Failed",
    "413": "Content Too Large",
    "414": "URI Too Long",
    "415": "Unsupported Media Type",
    "416": "Range Not Satisfiable",
    "417": "Expectation Failed",
    "418": "I'm a Teapot",
    "421": "Misdirected Request",
    "422": "Unprocessable Content",
    "423": "Locked",
    "424": "Failed Dependency",
    "425": "Too Early",
    "426": "Upgrade Required",
    "427": "Unassigned: ",
    "428": "Precondition Required",
    "429": "Too Many Requests",
    "431": "Request Header Fields Too Large",
    "451": "Unavailable For Legal Reasons",
    "500": "Internal Server Error",
    "501": "Not Implemented",
    "502": "Bad Gateway",
    "503": "Service Unavailable",
    "504": "Gateway Timeout",
    "505": "HTTP Version Not Supported",
    "506": "Variant Also Negotiates",
    "507": "Insufficient Storage",
    "508": "Loop Detected",
    "509": "Unassigned: ",
    "510": "Not Extended",
    "511": "Network Authentication Required",
}


def pretty(response, content=False):
    RESET = "\033[0m"
    HIGHLIGHT = "\033[34m"

    # status code
    print(HIGHLIGHT + str(response.status) + " " + RESET + STATUS_CODES[str(response.status)])

    # headers
    for k in sorted(response.headers.keys()):
        print(HIGHLIGHT + k + RESET + ": " + response.headers[k])

    if content:
        # blank line
        print()

        # response body
        if response.json:
            print(json_lib.dumps(response.json, indent=2))
        else:
            print(response.content.decode())



import datetime
import json
import os
import sys
import urllib

from importlib import resources
from . import checks as checks_module

VERSION = "1.2.7"

from ready.checks.bad_response import (
    check_bad_response_cloudflare,
    check_bad_response_kasada,
)
from ready.checks.content import (
    check_http_cache_control_is_included,
    check_http_content_type_header_contains_charset,
    check_http_expires_header_is_not_set,
    check_http_p3p_header_is_not_set,
    check_http_response_should_be_gzipped,
    check_http_response_should_include_content_type,
)
from ready.checks.cookies import (
    check_cookies_should_be_httponly,
    check_cookies_should_be_samesite,
    check_cookies_should_be_secure,
)
from ready.checks.corp_coop_coep import (
    check_cross_origin_embedder_policy_should_be_require_corp,
    check_cross_origin_opener_policy_should_be_sameorigin,
    check_cross_origin_resource_policy_should_be_sameorigin,
)
from ready.checks.csp import (
    check_csp_includes_default_or_script_directive,
    check_csp_must_not_include_report_sample,
    check_csp_must_not_include_unsafe_eval,
    check_csp_must_not_include_unsafe_inline,
    check_csp_should_exist,
    check_csp_should_include_reportto,
    check_csp_must_not_include_reporturi,
    check_csp_should_only_include_valid_directives,
    check_csp_should_start_with_defaultsrc_none,
    check_csp_upgrade_insecure_requests,
)
from ready.checks.dns import check_aaaa_record_exists
from ready.checks.email import (
    check_dmarc_record_should_exist,
    check_dmarc_record_should_reject_failures,
    check_spf_dash_all,
    check_spf_dns_record_does_not_exist,
    check_spf_record_should_exist,
    check_spf_txt_record_should_disallow_all,
    check_spf_uses_less_than_10_requests,
)
from ready.checks.hsts import (
    check_hsts_header_should_be_included_in_response,
    check_hsts_header_should_have_a_long_max_age,
    check_hsts_header_should_have_includesubdomains,
    check_hsts_header_should_have_preload,
)
from ready.checks.html import (
    check_cdns_should_not_be_used,
    check_frame_ancestors_should_exist,
    check_html_includes_rel_icon,
    check_html_includes_title,
    check_html_meta_charset,
    check_html_script_tags_use_sri,
    check_html_should_not_be_cached_for_more_than_24_hours,
    check_html_should_not_use_schemeless_urls,
    check_html_should_not_use_unnecessary_entities,
    check_html_starts_with_doctype,
    check_html_tag_includes_lang,
    check_permissions_policy_should_exist,
    check_referrer_policy_should_be_set,
    check_rss_should_return_cors_header,
    check_x_content_type_options_should_be_nosniff,
    check_x_dns_prefetch_control_is_off,
    check_x_xss_protection_should_not_exist,
)
from ready.checks.leaky_headers import check_should_not_include_leaky_headers
from ready.checks.ns import check_at_least_two_nameservers_configured
from ready.checks.redirect import check_http_to_https_redirect
from ready.checks.report_to import check_report_to_header_must_not_be_included_in_response
from ready.checks.ssl import (
    check_dns_caa_record_should_exist,
    check_dns_caa_record_should_include_accounturi,
    check_dns_caa_record_should_include_validationmethods,
    check_ssl_certificate_should_be_trusted,
    check_ssl_connection_fails_with_tls_1_0,
    check_ssl_connection_fails_with_tls_1_1,
    check_ssl_expiry_should_be_greater_than_five_days,
    check_ssl_expiry_should_be_less_than_one_year,
)
from ready.checks.status import check_http_response_should_be_200
from ready.checks.swagger import check_swagger_should_not_return_200
from ready.checks.well_known import (
    check_favicon_is_served,
    check_robots_txt_exists,
    check_security_txt_exists,
    check_security_txt_not_expired,
)
from ready.thttp import pretty, request

USE_FLD = True

try:
    from tld import get_fld
except ImportError:
    USE_FLD = False


DEFAULT_HEADERS = {
    # MacOS Safari
    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1 Safari/605.1.15",
    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "accept-language": "en-AU,en;q=0.9",
    "accept-encoding": "gzip",
}


def response_or_none(url, name="", request_filter="", **kwargs):
    if request_filter and request_filter not in name:
        print(f"Skipping HTTP request {name}")
        return None

    try:
        response = request(url, **kwargs)
        return response
    except urllib.error.URLError:
        return None
    except Exception as e:
        print(url, type(e))
        return None


def ready(
    domain,
    print_headers=False,
    print_content=False,
    json_output=False,
    hide_output=False,
    fuzz=False,
    check_filter=None,
    request_filter=None,
    extra_args={},
):
    domain_with_no_path = urllib.parse.urlparse("https://" + domain).hostname

    if USE_FLD:
        fld = get_fld(domain, fix_protocol=True)
    else:
        fld = "Disabled. Install tld if fld is different to domain."

    if not hide_output:
        print(f"Domain: {domain}, Domain (no path): {domain_with_no_path}, First Level Domain: {fld}")

    responses = {
        "http_response": response_or_none(
            f"http://{domain}", "http_response", request_filter, verify=False, headers=DEFAULT_HEADERS, timeout=3
        ),
        "response": response_or_none(
            f"https://{domain}", "response", request_filter, verify=False, headers=DEFAULT_HEADERS, timeout=3
        ),
    }

    if not responses["response"]:
        print(f"No response from https://{domain}")

        if request_filter:
            print("Request filter in place, continuing...")
        else:
            return None

    responses["security_txt_response"] = response_or_none(
        f"https://{domain_with_no_path}/.well-known/security.txt",
        "security_txt_response",
        request_filter,
        headers=DEFAULT_HEADERS,
        timeout=3,
    )

    responses["robots_txt_response"] = response_or_none(
        f"https://{domain_with_no_path}/robots.txt", "robots_txt_response", request_filter, headers=DEFAULT_HEADERS, timeout=3
    )

    responses["favicon_response"] = response_or_none(
        f"https://{domain_with_no_path}/favicon.ico",
        "favicon_response",
        request_filter,
        verify=False,
        headers=DEFAULT_HEADERS,
        timeout=3,
    )

    responses["dns_ns_response"] = response_or_none(
        f"https://dns.google/resolve?name={domain_with_no_path}&type=NS", "dns_ns_response", request_filter
    )
    responses["dns_mx_response"] = response_or_none(
        f"https://dns.google/resolve?name={domain_with_no_path}&type=MX", "dns_mx_response", request_filter
    )
    responses["dns_txt_response"] = response_or_none(
        f"https://dns.google/resolve?name={domain_with_no_path}&type=TXT", "dns_txt_response", request_filter
    )
    responses["dns_spf_response"] = response_or_none(
        f"https://dns.google/resolve?name={domain_with_no_path}&type=SPF", "dns_spf_response", request_filter
    )
    responses["dns_caa_response"] = response_or_none(
        f"https://dns.google/resolve?name={domain_with_no_path}&type=CAA", "dns_caa_response", request_filter
    )
    responses["dns_a_response"] = response_or_none(
        f"https://dns.google/resolve?name={domain_with_no_path}&type=A", "dns_aaaa_response", request_filter
    )
    responses["dns_aaaa_response"] = response_or_none(
        f"https://dns.google/resolve?name={domain_with_no_path}&type=AAAA", "dns_aaaa_response", request_filter
    )
    responses["dns_dmarc_response"] = response_or_none(
        f"https://dns.google/resolve?name=_dmarc.{domain_with_no_path}&type=TXT", "dns_dmarc_response", request_filter
    )

    if USE_FLD and domain != fld:
        responses["dns_ns_response_fld"] = response_or_none(f"https://dns.google/resolve?name={fld}&type=NS")
        responses["dns_mx_response_fld"] = response_or_none(f"https://dns.google/resolve?name={fld}&type=MX")
        responses["dns_spf_response_fld"] = response_or_none(f"https://dns.google/resolve?name={fld}&type=SPF")
        responses["dns_txt_response_fld"] = response_or_none(f"https://dns.google/resolve?name={fld}&type=TXT")
        responses["dns_dmarc_response_fld"] = response_or_none(f"https://dns.google/resolve?name=_dmarc.{fld}&type=TXT")
        responses["dns_caa_response_fld"] = response_or_none(f"https://dns.google/resolve?name={fld}&type=CAA")

    checks = []
    is_html = responses["response"] and "html" in responses["response"].headers.get("content-type", "")

    a_records = [x["data"] for x in responses["dns_a_response"].json.get("Answer", [])]
    aaaa_records = [x["data"] for x in responses["dns_aaaa_response"].json.get("Answer", [])]
    extra_args["is_ipv6"] = len(a_records) == 0 and len(aaaa_records) > 0

    # TODO: accept argument to _not_ print to stdout
    if print_headers:
        pretty(responses["response"], content=False)
        print()

    if print_content:
        print(responses["response"].content)

    # bad response checks go first
    checks = [
        check_bad_response_kasada,
        check_bad_response_cloudflare,
    ]

    checks.extend(
        [
            check_http_to_https_redirect,
            check_http_response_should_be_200,
            check_http_response_should_include_content_type,
            check_aaaa_record_exists,
            check_hsts_header_should_be_included_in_response,
            check_hsts_header_should_have_a_long_max_age,
            check_hsts_header_should_have_includesubdomains,
            check_hsts_header_should_have_preload,
            check_csp_should_exist,
            check_csp_should_start_with_defaultsrc_none,
            check_csp_includes_default_or_script_directive,
            check_csp_must_not_include_unsafe_inline,
            check_csp_must_not_include_unsafe_eval,
            check_csp_must_not_include_report_sample,
            check_csp_upgrade_insecure_requests,
            check_csp_must_not_include_reporturi,
            check_csp_should_include_reportto,
            check_csp_should_only_include_valid_directives,
            check_report_to_header_must_not_be_included_in_response,
            check_robots_txt_exists,
            check_security_txt_exists,
            check_security_txt_not_expired,
            check_favicon_is_served,
            check_http_response_should_be_gzipped,
            check_http_content_type_header_contains_charset,
            check_http_expires_header_is_not_set,
            check_http_cache_control_is_included,
            check_http_p3p_header_is_not_set,
            check_referrer_policy_should_be_set,
            check_cross_origin_resource_policy_should_be_sameorigin,
            check_cross_origin_opener_policy_should_be_sameorigin,
            check_cross_origin_embedder_policy_should_be_require_corp,
            check_should_not_include_leaky_headers,
            check_ssl_expiry_should_be_less_than_one_year,
            check_ssl_expiry_should_be_greater_than_five_days,
            check_ssl_certificate_should_be_trusted,
            check_ssl_connection_fails_with_tls_1_1,
            check_ssl_connection_fails_with_tls_1_0,
            check_dns_caa_record_should_exist,
            check_dns_caa_record_should_include_accounturi,
            check_dns_caa_record_should_include_validationmethods,
            check_at_least_two_nameservers_configured,
            check_cookies_should_be_samesite,
            check_cookies_should_be_secure,
            check_cookies_should_be_httponly,
            check_spf_dash_all,
            check_spf_record_should_exist,
            check_spf_dns_record_does_not_exist,
            check_spf_txt_record_should_disallow_all,
            check_dmarc_record_should_exist,
            check_dmarc_record_should_reject_failures,
            check_spf_uses_less_than_10_requests,
        ]
    )

    if is_html:
        checks.extend(
            [
                check_permissions_policy_should_exist,
                check_frame_ancestors_should_exist,
                check_x_content_type_options_should_be_nosniff,
                check_x_xss_protection_should_not_exist,
                check_html_starts_with_doctype,
                check_html_tag_includes_lang,
                check_html_meta_charset,
                check_html_includes_title,
                check_html_includes_rel_icon,
                check_html_should_not_use_schemeless_urls,
                check_html_script_tags_use_sri,
                check_html_should_not_use_unnecessary_entities,
                check_html_should_not_be_cached_for_more_than_24_hours,
                check_x_dns_prefetch_control_is_off,
                check_cdns_should_not_be_used,
                check_rss_should_return_cors_header,
            ]
        )

    if fuzz:
        checks.extend(
            [
                check_swagger_should_not_return_200,
            ]
        )

    extra_args["print_output"] = not hide_output

    results = []
    for c in checks:
        if check_filter and check_filter not in c.__name__:
            continue

        result = c(responses, domain=domain, domain_with_no_path=domain_with_no_path, **extra_args)
        if result:
            results.append(result)

    if json_output:
        print(
            json.dumps(
                {
                    "domain": domain,
                    "score": score_from_results(results),
                    "checks": {
                        r.check: {
                            "passed": r.passed,
                            "message": r.message,
                        }
                        for r in results
                    },
                    "when": datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "version": VERSION,
                },
                indent=2,
            )
        )

    return results


def score_from_results(results):
    return 100 - 3 * len([x for x in results if not x.passed and not x.warn_on_fail])


def parse_args(args):
    result = {
        a.split("=")[0]: (
            int(a.split("=")[1]) if "=" in a and a.split("=")[1].isnumeric() else a.split("=")[1] if "=" in a else True
        )
        for a in args
        if "--" in a
    }
    result["[]"] = [a for a in args if not a.startswith("--")]
    return result


def usage():
    print("ready")
    print("")
    print("Usage: ready.py [--headers] [--content] [--json] [--quiet] [--score] [--fuzz] <domain>")
    print("")
    print("  --headers      Output the headers from the HTTPS request made to the domain")
    print("  --content      Output the content from the HTTPS request made to the domain")
    print("  --fuzz         Include checks that fuzz urls (only run this on your own domain)")
    print("  --json         Provide JSON output")
    print("  --quiet        No text output")
    print("  --score        Print a score out of 100 for this domain")
    print("  --doc          Print the list of check names")
    print("  --version      Print version information")

    print("\nDevelopment / experimental options for filtering checks and HTTP requests during testing:")
    print("")
    print("  --check-filter=<x>     Only run checks that match the provided filter")
    print("  --request-filter=<x>   Only make HTTP requests that match the provided filter")


def cli():
    args = parse_args(sys.argv[1:])

    if "--version" in args:
        print(f"ready {VERSION}")
        sys.exit()

    if "--doc" in args:
        for f in resources.files(checks_module).iterdir():
            if f.name.endswith(".py"):
                for line in open(f).readlines():
                    if line.strip().startswith("# Check: "):
                        print(line.strip().replace("# Check: ", "- "))
        sys.exit()

    if "--help" in args or not args["[]"]:
        usage()
        sys.exit()

    results = ready(
        args["[]"][0],
        print_headers=args.get("--headers", False),
        print_content=args.get("--content", False),
        json_output=args.get("--json", False),
        hide_output=args.get("--quiet", False),
        fuzz=args.get("--fuzz", False),
        check_filter=args.get("--check-filter", ""),
        request_filter=args.get("--request-filter", ""),
    )

    if "--score" in args:
        print(f"Score: {score_from_results(results)}/100")


if __name__ == "__main__":
    cli()



from ready.result import result
from datetime import datetime, timezone


# Check: Robots.txt exists and is a text file
def check_robots_txt_exists(responses, **kwargs):
    robots_response = responses["robots_txt_response"]

    return result(
        robots_response and robots_response.status == 200 and "text/plain" in robots_response.headers.get("content-type", ""),
        "Robots.txt exists and is a text file",
        "wellknown_robots",
        **kwargs,
    )


# Check: Security.txt exists and is a text file that contains required attributes
def check_security_txt_exists(responses, **kwargs):
    security_txt_response = responses["security_txt_response"]

    return result(
        security_txt_response
        and security_txt_response.status == 200
        and "text/plain" in security_txt_response.headers.get("content-type", "")
        and b"Contact:" in security_txt_response.content
        and b"Expires:" in security_txt_response.content,
        "Security.txt exists and is a text file that contains required attributes",
        "wellknown_security",
        **kwargs,
    )


# Check: Security.txt has an expiry date in the future
def check_security_txt_not_expired(responses, **kwargs):
    security_txt_response = responses["security_txt_response"]

    for line in security_txt_response.content.splitlines():
        line = line.decode()

        if line.strip().startswith("Expires:"):
            date = line.replace("Expires:", "").strip()

            try:
                dt = datetime.fromisoformat(date.upper())

                return result(
                    dt > datetime.utcnow().replace(tzinfo=timezone.utc),
                    f"Security.txt has an expiry date in the future ({dt})",
                    "wellknown_security_not_expired",
                    **kwargs,
                )
            except:
                break

    return result(
        False,
        "Security.txt has an expiry date in the future (missing file or expires line)",
        "wellknown_security_not_expired",
        **kwargs,
    )


# Check: Favicon is served at /favicon.ico
def check_favicon_is_served(responses, **kwargs):
    favicon_response = responses["favicon_response"]
    return result(
        favicon_response.status == 200 and favicon_response.headers.get("content-type", "").startswith("image/"),
        "Favicon is served at /favicon.ico",
        "wellknown_favicon",
        **kwargs,
    )



import errno
import socket
import ssl
from datetime import date, datetime

from ready.result import result
from ready.thttp import request

CONNECTION_TIMEOUT = 5.0


class SSLConnectionFailed(Exception):
    pass


class UnknownSSLFailure(Exception):
    pass


class LookupFailed(Exception):
    pass


def connect_with_specific_protocol(domain, protocol, ipv6=False):
    try:
        sock_type = socket.AF_INET6 if ipv6 else socket.AF_INET
        sock = socket.socket(sock_type, socket.SOCK_STREAM)

        context = ssl.SSLContext(protocol=protocol)
        ssl_sock = context.wrap_socket(sock, server_hostname=domain)
        ssl_sock.settimeout(CONNECTION_TIMEOUT)
        ssl_sock.connect((domain, 443))
        ssl_sock.close()
        return True
    except:
        return False


def get_ssl_expiry(domain, ipv6=False):
    try:
        sock_type = socket.AF_INET6 if ipv6 else socket.AF_INET
        sock = socket.socket(sock_type, socket.SOCK_STREAM)
        sock.settimeout(CONNECTION_TIMEOUT)

        context = ssl.create_default_context()
        ssl_sock = context.wrap_socket(sock, server_hostname=domain)
        ssl_sock.settimeout(CONNECTION_TIMEOUT)
        ssl_sock.connect((domain, 443))

        cert = ssl_sock.getpeercert()
        end = datetime.fromtimestamp(ssl.cert_time_to_seconds(cert["notAfter"]))
        ssl_sock.close()
        return end.date()
    except:
        return None


# Check: SSL certificate should be trusted
def check_ssl_certificate_should_be_trusted(responses, **kwargs):
    try:
        response = request(f'https://{kwargs["domain"]}', verify=True)
        return result(
            True,
            f"SSL certificate should be trusted",
            "ssl_trusted",
            **kwargs,
        )
    except:
        return result(False, f"SSL certificate should be trusted", "ssl_trusted", **kwargs)


# Check: SSL expiry should be less than one year
def check_ssl_expiry_should_be_less_than_one_year(responses, **kwargs):
    ssl_expiry = get_ssl_expiry(kwargs["domain_with_no_path"], ipv6=kwargs["is_ipv6"])
    ssl_expiry_days = (ssl_expiry - date.today()).days if ssl_expiry else None

    return result(
        ssl_expiry_days and ssl_expiry_days < 366,
        f"SSL expiry should be less than one year ({ssl_expiry_days} days)",
        "ssl_expiry_max",
        **kwargs,
    )


# Check: SSL expiry should be greater than five days
def check_ssl_expiry_should_be_greater_than_five_days(responses, **kwargs):
    ssl_expiry = get_ssl_expiry(kwargs["domain_with_no_path"], ipv6=kwargs["is_ipv6"])
    ssl_expiry_days = (ssl_expiry - date.today()).days if ssl_expiry else None

    return result(
        ssl_expiry_days and ssl_expiry_days > 5,
        f"SSL expiry should be greater than five days ({ssl_expiry_days} days)",
        "ssl_expiry_min",
        **kwargs,
    )


# Check: SSL connection fails when using TLS 1.1
def check_ssl_connection_fails_with_tls_1_1(responses, **kwargs):
    domain = kwargs["domain"]
    connection_successful = connect_with_specific_protocol(domain, ssl.PROTOCOL_TLSv1_1, ipv6=kwargs["is_ipv6"])

    return result(
        not connection_successful,
        f"SSL connection fails when using TLS 1.1",
        "ssl_tls_1_1",
        **kwargs,
    )


# Check: SSL connection fails when using TLS 1.0
def check_ssl_connection_fails_with_tls_1_0(responses, **kwargs):
    domain = kwargs["domain"]
    connection_successful = connect_with_specific_protocol(domain, ssl.PROTOCOL_TLSv1, ipv6=kwargs["is_ipv6"])

    return result(
        not connection_successful,
        f"SSL connection fails when using TLS 1.0",
        "ssl_tls_1_0",
        **kwargs,
    )


# Check: DNS CAA should be enabled
# https://blog.qualys.com/product-tech/2017/03/13/caa-mandated-by-cabrowser-forum
def check_dns_caa_record_should_exist(responses, **kwargs):
    records = [
        r["data"] for r in responses["dns_caa_response"].json.get("Answer", []) if "data" in r and r.get("type", 0) == 257
    ]

    if not records and "dns_caa_response_fld" in responses:
        records = [
            r["data"]
            for r in responses["dns_caa_response_fld"].json.get("Answer", [])
            if "data" in r and r.get("type", 0) == 257
        ]

    return result(
        records and all(["issue" in r or "iodef" in r for r in records]),
        f"DNS CAA should be enabled ({records})",
        "ssl_dns_caa",
        **kwargs,
    )


# Check: DNS CAA should include accounturi
def check_dns_caa_record_should_include_accounturi(responses, **kwargs):
    records = [
        r["data"] for r in responses["dns_caa_response"].json.get("Answer", []) if "data" in r and r.get("type", 0) == 257
    ]

    if not records and "dns_caa_response_fld" in responses:
        records = [
            r["data"]
            for r in responses["dns_caa_response_fld"].json.get("Answer", [])
            if "data" in r and r.get("type", 0) == 257
        ]

    # filter to just the issue records
    records = [r for r in records if "issue " in r]

    return result(
        records and all(["accounturi=" in r for r in records]),
        f"DNS CAA should include accounturi ({records})",
        "ssl_dns_caa_accounturi",
        warn_on_fail=True,
        **kwargs,
    )


# Check: DNS CAA should include validationmethods
def check_dns_caa_record_should_include_validationmethods(responses, **kwargs):
    records = [
        r["data"] for r in responses["dns_caa_response"].json.get("Answer", []) if "data" in r and r.get("type", 0) == 257
    ]

    if not records and "dns_caa_response_fld" in responses:
        records = [
            r["data"]
            for r in responses["dns_caa_response_fld"].json.get("Answer", [])
            if "data" in r and r.get("type", 0) == 257
        ]

    # filter to just the issue records
    records = [r for r in records if "issue " in r]

    return result(
        records and all(["validationmethods=" in r for r in records]),
        f"DNS CAA should include validationmethods ({records})",
        "ssl_dns_caa_validationmethods",
        warn_on_fail=True,
        **kwargs,
    )



from ready.result import result

USE_BS4 = True

try:
    from bs4 import BeautifulSoup
except ImportError:
    USE_BS4 = False


def extract_csp(response):
    if "content-security-policy" in response.headers:
        return response.headers["content-security-policy"]

    if USE_BS4:
        soup = BeautifulSoup(response.content, "html.parser")
        meta_tags = soup.find_all("meta")
        for t in meta_tags:
            if t.attrs.get("http-equiv", "").lower() == "content-security-policy":
                return t.attrs.get("content", "")
    else:
        print("No Content-Security-Policy header, and beautifulsoup4 is not installed to inspect HTML")

    return None


def _trunc(s, max_length=200):
    if not s:
        return ""

    if len(s) > max_length:
        return s[:max_length] + "..."

    return s


# Check: Content-Security-Policy header should exist
def check_csp_should_exist(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp != None,
        f"Content-Security-Policy header should exist ({_trunc(csp)})",
        "csp",
        **kwargs,
    )


# Check: Content-Security-Policy header should start with default-src 'none'
def check_csp_should_start_with_defaultsrc_none(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp != None and csp.startswith("default-src 'none'"),
        f"Content-Security-Policy header should start with default-src 'none' ({_trunc(csp)})",
        "csp_defaultsrc_none",
        **kwargs,
    )


# Check: Content-Security-Policy must include either default-src or script-src
def check_csp_includes_default_or_script_directive(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp != None and ("default-src" in csp or "script-src" in csp),
        f"Content-Security-Policy must include either default-src or script-src ({_trunc(csp)})",
        "csp_required_directives",
        **kwargs,
    )


# Check: Content-Security-Policy header must not include unsafe-inline
# NOTE: this checks everywhere, not just in script-src
def check_csp_must_not_include_unsafe_eval(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp != None and "unsafe-eval" not in csp,
        f"Content-Security-Policy header must not include unsafe-eval ({_trunc(csp)})",
        "csp_no_unsafe_inline",
        **kwargs,
    )


# Check: Content-Security-Policy header must not include unsafe-eval
def check_csp_must_not_include_unsafe_inline(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp != None and "unsafe-inline" not in csp,
        f"Content-Security-Policy header must not include unsafe-inline ({_trunc(csp)})",
        "csp_no_unsafe_inline",
        **kwargs,
    )


# Check: Content-Security-Policy header must not include report-sample
def check_csp_must_not_include_report_sample(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp != None and "script-sample" not in csp,
        f"Content-Security-Policy header must not include report-sample ({_trunc(csp)})",
        "csp_no_report_sample",
        **kwargs,
    )


# Check: Content-Security-Policy header must not include report-uri
# NOTE: report-uri is being replaced by report-to but browser support is spotty so report-uri should still exist
def check_csp_must_not_include_reporturi(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp is None or (("report-uri https://" not in csp)),
        f"Content-Security-Policy header must not include report-uri ({_trunc(csp)})",
        "csp_report_uri",
        warn_on_fail=False,
        **kwargs,
    )


# Check: Content-Security-Policy header should not include report-to
def check_csp_should_include_reportto(responses, **kwargs):
    csp = extract_csp(responses["response"])

    if not csp:
        csp = ""

    return result(
        ("report-to" not in csp),
        f"Content-Security-Policy header should not include report-to ({_trunc(csp)})",
        "csp_report_to",
        warn_on_fail=True,
        **kwargs,
    )


# Check: Content-Security-Policy header should include upgrade-insecure-requests
def check_csp_upgrade_insecure_requests(responses, **kwargs):
    csp = extract_csp(responses["response"])

    return result(
        csp != None and "upgrade-insecure-requests" in csp,
        f"Content-Security-Policy header should include upgrade-insecure-requests ({_trunc(csp)})",
        "csp_upgrade_insecure_requests",
        **kwargs,
    )


# Check: Content-Security-Policy header only includes valid directives
def check_csp_should_only_include_valid_directives(responses, **kwargs):
    csp = extract_csp(responses["response"])

    directives = []

    if csp:
        for directive in csp.split(";"):
            directive = directive.strip()

            if " " in directive:
                directives.append(directive.split()[0])
            else:
                directives.append(directive.strip())

    valid_directives = [
        "base-uri",
        "block-all-mixed-content",
        "child-src",
        "connect-src",
        "default-src",
        "font-src",
        "form-action",
        "frame-ancestors",
        "frame-src",
        "img-src",
        "manifest-src",
        "media-src",
        "navigate-to",
        "object-src",
        "plugin-types",
        "prefetch-src",
        "report-to",
        "report-uri",
        "require-sri-for",
        "require-trusted-types-for",
        "sandbox",
        "script-src-attr",
        "script-src-elem",
        "script-src",
        "style-src-attr",
        "style-src-elem",
        "style-src",
        "trusted-types",
        "upgrade-insecure-requests",
        "worker-src",
    ]

    return result(
        csp != None and all([x in valid_directives for x in directives]),
        f"Content-Security-Policy header only includes valid directives ({directives})",
        "csp_valid_directives",
        warn_on_fail=False,
        **kwargs,
    )



import re

from ready.result import result


# Check: Report-To Header must not be included in response
def check_report_to_header_must_not_be_included_in_response(responses, **kwargs):
    return result(
        responses["response"].headers.get("report-to") in [None, ""],
        f"Report-To Header must not be included in response ({responses['response'].headers.get('report-to')})",
        "report_to",
        **kwargs,
    )



from ready.result import result


# Check: Cookies should set the SameSite flag
def check_cookies_should_be_samesite(responses, **kwargs):
    cookies = responses["response"].headers.get("set-cookie", "")
    cookie_note = cookies or "no cookie set"

    return result(
        not cookies or "samesite=" in cookies.lower(),
        f"Cookies should set the SameSite flag ({cookie_note})",
        "cookies_samesite",
        **kwargs,
    )


# Check: Cookies should set the Secure flag
def check_cookies_should_be_secure(responses, **kwargs):
    cookies = responses["response"].headers.get("set-cookie", "")
    cookie_note = cookies or "no cookie set"

    return result(
        not cookies or "secure;" in cookies.lower(),
        f"Cookies should set the Secure flag ({cookie_note})",
        "cookies_secure",
        **kwargs,
    )


# Check: Cookies should set the HttpOnly flag
def check_cookies_should_be_httponly(responses, **kwargs):
    cookies = responses["response"].headers.get("set-cookie", "")
    cookie_note = cookies or "no cookie set"

    return result(
        not cookies or "httponly;" in cookies.lower(),
        f"Cookies should set the HttpOnly flag ({cookie_note})",
        "cookies_httponly",
        **kwargs,
    )
